# Copyright 2020 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Model Builder for EfficientNet Edge Models.

efficientnet-litex (x=0,1,2,3,4) checkpoints are located in:
  https://storage.googleapis.com/cloud-tpu-checkpoints/efficientnet/lite/efficientnet-litex.tar.gz
"""
import os
from absl import logging
import tensorflow.compat.v1 as tf

from tensorflow_examples.lite.model_maker.third_party.efficientdet import utils
from tensorflow_examples.lite.model_maker.third_party.efficientdet.backbone import efficientnet_builder
from tensorflow_examples.lite.model_maker.third_party.efficientdet.backbone import efficientnet_model

# Edge models use inception-style MEAN and STDDEV for better post-quantization.
MEAN_RGB = [127.0, 127.0, 127.0]
STDDEV_RGB = [128.0, 128.0, 128.0]


def efficientnet_lite_params(model_name):
  """Get efficientnet params based on model name."""
  params_dict = {
      # (width_coefficient, depth_coefficient, resolution, dropout_rate)
      'efficientnet-lite0': (1.0, 1.0, 224, 0.2),
      'efficientnet-lite1': (1.0, 1.1, 240, 0.2),
      'efficientnet-lite2': (1.1, 1.2, 260, 0.3),
      'efficientnet-lite3': (1.2, 1.4, 280, 0.3),
      'efficientnet-lite4': (1.4, 1.8, 300, 0.3),
  }
  return params_dict[model_name]


_DEFAULT_BLOCKS_ARGS = [
    'r1_k3_s11_e1_i32_o16_se0.25', 'r2_k3_s22_e6_i16_o24_se0.25',
    'r2_k5_s22_e6_i24_o40_se0.25', 'r3_k3_s22_e6_i40_o80_se0.25',
    'r3_k5_s11_e6_i80_o112_se0.25', 'r4_k5_s22_e6_i112_o192_se0.25',
    'r1_k3_s11_e6_i192_o320_se0.25',
]


def efficientnet_lite(width_coefficient=None,
                      depth_coefficient=None,
                      dropout_rate=0.2,
                      survival_prob=0.8):
  """Creates a efficientnet model."""
  global_params = efficientnet_model.GlobalParams(
      blocks_args=_DEFAULT_BLOCKS_ARGS,
      batch_norm_momentum=0.99,
      batch_norm_epsilon=1e-3,
      dropout_rate=dropout_rate,
      survival_prob=survival_prob,
      data_format='channels_last',
      num_classes=1000,
      width_coefficient=width_coefficient,
      depth_coefficient=depth_coefficient,
      depth_divisor=8,
      min_depth=None,
      relu_fn=tf.nn.relu6,  # Relu6 is for easier quantization.
      # The default is TPU-specific batch norm.
      # The alternative is tf.layers.BatchNormalization.
      batch_norm=utils.TpuBatchNormalization,  # TPU-specific requirement.
      clip_projection_output=False,
      fix_head_stem=True,  # Don't scale stem and head.
      local_pooling=True,  # special cases for tflite issues.
      use_se=False)  # SE is not well supported on many lite devices.
  return global_params


def get_model_params(model_name, override_params):
  """Get the block args and global params for a given model."""
  if model_name.startswith('efficientnet-lite'):
    width_coefficient, depth_coefficient, _, dropout_rate = (
        efficientnet_lite_params(model_name))
    global_params = efficientnet_lite(
        width_coefficient, depth_coefficient, dropout_rate)
  else:
    raise NotImplementedError('model name is not pre-defined: %s' % model_name)

  if override_params:
    # ValueError will be raised here if override_params has fields not included
    # in global_params.
    global_params = global_params._replace(**override_params)

  decoder = efficientnet_builder.BlockDecoder()
  blocks_args = decoder.decode(global_params.blocks_args)

  logging.info('global_params= %s', global_params)
  return blocks_args, global_params


def build_model(images,
                model_name,
                training,
                override_params=None,
                model_dir=None,
                fine_tuning=False,
                features_only=False,
                pooled_features_only=False):
  """A helper function to create a model and return predicted logits.

  Args:
    images: input images tensor.
    model_name: string, the predefined model name.
    training: boolean, whether the model is constructed for training.
    override_params: A dictionary of params for overriding. Fields must exist in
      efficientnet_model.GlobalParams.
    model_dir: string, optional model dir for saving configs.
    fine_tuning: boolean, whether the model is used for finetuning.
    features_only: build the base feature network only (excluding final
      1x1 conv layer, global pooling, dropout and fc head).
    pooled_features_only: build the base network for features extraction (after
      1x1 conv layer and global pooling, but before dropout and fc head).

  Returns:
    logits: the logits tensor of classes.
    endpoints: the endpoints for each layer.

  Raises:
    When model_name specified an undefined model, raises NotImplementedError.
    When override_params has invalid fields, raises ValueError.
  """
  assert isinstance(images, tf.Tensor)
  assert not (features_only and pooled_features_only)

  # For backward compatibility.
  if override_params and override_params.get('drop_connect_rate', None):
    override_params['survival_prob'] = 1 - override_params['drop_connect_rate']

  if not training or fine_tuning:
    if not override_params:
      override_params = {}
    override_params['batch_norm'] = utils.BatchNormalization
  blocks_args, global_params = get_model_params(model_name, override_params)

  if model_dir:
    param_file = os.path.join(model_dir, 'model_params.txt')
    if not tf.gfile.Exists(param_file):
      if not tf.gfile.Exists(model_dir):
        tf.gfile.MakeDirs(model_dir)
      with tf.gfile.GFile(param_file, 'w') as f:
        logging.info('writing to %s', param_file)
        f.write('model_name= %s\n\n' % model_name)
        f.write('global_params= %s\n\n' % str(global_params))
        f.write('blocks_args= %s\n\n' % str(blocks_args))

  model = efficientnet_model.Model(blocks_args, global_params, model_name)
  outputs = model(
      images,
      training=training,
      features_only=features_only,
      pooled_features_only=pooled_features_only)
  features, endpoints = outputs[0], outputs[1:]
  if features_only:
    features = tf.identity(features, 'features')
  elif pooled_features_only:
    features = tf.identity(features, 'pooled_features')
  else:
    features = tf.identity(features, 'logits')
  return features, endpoints


def build_model_base(images, model_name, training, override_params=None):
  """Create a base feature network and return the features before pooling.

  Args:
    images: input images tensor.
    model_name: string, the predefined model name.
    training: boolean, whether the model is constructed for training.
    override_params: A dictionary of params for overriding. Fields must exist in
      efficientnet_model.GlobalParams.

  Returns:
    features: base features before pooling.
    endpoints: the endpoints for each layer.

  Raises:
    When model_name specified an undefined model, raises NotImplementedError.
    When override_params has invalid fields, raises ValueError.
  """
  assert isinstance(images, tf.Tensor)
  # For backward compatibility.
  if override_params and override_params.get('drop_connect_rate', None):
    override_params['survival_prob'] = 1 - override_params['drop_connect_rate']

  blocks_args, global_params = get_model_params(model_name, override_params)

  model = efficientnet_model.Model(blocks_args, global_params, model_name)
  outputs = model(images, training=training, features_only=True)

  return outputs[0], outputs[1:]



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Audio model specification."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc
import collections
import csv
import io
import os
import tempfile

import tensorflow as tf
from tensorflow_examples.lite.model_maker.core.api.api_util import mm_export
from tensorflow_examples.lite.model_maker.core.task import model_util
import tensorflow_hub as hub

try:
  from tflite_support.metadata_writers import audio_classifier as md_writer  # pylint: disable=g-import-not-at-top
  from tflite_support.metadata_writers import metadata_info as md_info  # pylint: disable=g-import-not-at-top
  from tflite_support.metadata_writers import writer_utils  # pylint: disable=g-import-not-at-top
  ENABLE_METADATA = True
except ImportError:
  ENABLE_METADATA = False


class MetadataWriter:
  """Helper class to populate Audio Metadata, to be used in `with` statement.

  Simple usage for model with two classification heads.

  with MetadataWriter(tflite_path) as writer:
    writer.add_input(sample_rate=16000, channels=1)
    writer.add_output(name='animal_sound', labels=['dog', 'cat'])
    writer.add_output(name='speech_command', labels=['yes', 'no'])
    writer.save(tflite_path, json_filepath)

  `add_output` can also take an ordered dict for multiple locales, example:

  writer.add_output(name='animal_sound', labels=collections.OrderedDict([
    ('en', ['bird', 'cat']),
    ('fr', ['oiseau', 'chat'])
  ]))
  """

  def __init__(self, tflite_filepath, **kwargs):
    self._model = writer_utils.load_file(tflite_filepath)
    self._general_md = md_info.GeneralMd(**kwargs)
    self._inputs = []
    self._outputs = []

  def __enter__(self):
    self._temp_folder = tempfile.TemporaryDirectory()
    return self

  def __exit__(self, exc_type, exc_val, exc_tb):
    self._temp_folder.cleanup()
    # Delete the attribute so that it errors out if not in `with` statement.
    delattr(self, '_temp_folder')

  def add_input(self, **kwargs):
    """Add metadta for the input tensor."""
    self._inputs.append(md_info.InputAudioTensorMd(**kwargs))

  def add_output(self, name, labels, **kwargs):
    """Add metadata for output tensor in order."""
    if isinstance(labels, list):
      default_locale = None
      labels = collections.OrderedDict([(default_locale, labels)])
      return self.add_output(name, labels, **kwargs)

    label_files = []
    if isinstance(labels, collections.OrderedDict):
      for locale, label_list in labels.items():
        full_path = os.path.join(
            self._temp_folder.name,
            '{}_labels{}.txt'.format(name, '_' + locale if locale else ''))
        model_util.export_labels(full_path, label_list)
        label_files.append(
            md_info.LabelFileMd(file_path=full_path, locale=locale))
    else:
      raise ValueError(
          '`labels` should be either a list of labels or an ordered dict mapping `locale` -> list of labels. got: {}'
          .format(labels))

    idx = len(self._outputs)
    self._outputs.append(
        md_info.ClassificationTensorMd(
            name=name,
            label_files=label_files,
            tensor_type=writer_utils.get_output_tensor_types(self._model)[idx],
            **kwargs))

  def save(self, tflite_filepath=None, json_filepath=None):
    """Persist model with metadata."""
    if len(self._inputs) > 1:
      raise ValueError('Only supports single input, got {}'.format(
          len(self._inputs)))
    input_md = self._inputs[0]

    writer = md_writer.MetadataWriter.create_from_metadata_info_for_multihead(
        model_buffer=self._model,
        general_md=self._general_md,
        input_md=input_md,
        output_md_list=self._outputs)
    if tflite_filepath:
      writer_utils.save_file(writer.populate(), tflite_filepath, mode='wb')
    if json_filepath:
      writer_utils.save_file(
          writer.get_metadata_json(), json_filepath, mode='wt')


def _ensure_tf25(version):
  if version < '2.5':
    raise RuntimeError(
        'Audio Tasks requires TF2.5 or later. For example, you can run the '
        'following command to install TF2.5.0rc2:\n\n'
        'pip3 install tensorflow==2.5.0rc2\n\n')


def _get_tf_version():
  return tf.__version__


class BaseSpec(abc.ABC):
  """Base model spec for audio classification."""

  def __init__(self, model_dir=None, strategy=None):
    _ensure_tf25(_get_tf_version())
    self.model_dir = model_dir
    if not model_dir:
      self.model_dir = tempfile.mkdtemp()
    tf.compat.v1.logging.info('Checkpoints are stored in %s', self.model_dir)
    self.strategy = strategy or tf.distribute.get_strategy()

  @abc.abstractproperty
  def target_sample_rate(self):
    pass

  @abc.abstractmethod
  def create_model(self, num_classes, train_whole_model=False):
    pass

  @abc.abstractmethod
  def run_classifier(self, model, epochs, train_ds, validation_ds, **kwargs):
    pass

  def preprocess_ds(self, ds, is_training=False, cache_fn=None):
    """Returns a preprocessed dataset."""
    _ = is_training
    _ = cache_fn
    return ds

  def get_default_quantization_config(self):
    """Gets the default quantization configuration."""
    return None


def _remove_suffix_if_possible(text, suffix):
  return text.rsplit(suffix, 1)[0]


TFJS_MODEL_ROOT = 'https://storage.googleapis.com/tfjs-models/tfjs'


def _load_browser_fft_preprocess_model():
  """Load a model replicating WebAudio's AnalyzerNode.getFloatFrequencyData."""
  model_name = 'sc_preproc_model'
  file_extension = '.tar.gz'
  filename = model_name + file_extension
  # Load the preprocessing model, which transforms audio waveform into
  # spectrograms (2D image-like representation of sound).
  # This model replicates WebAudio's AnalyzerNode.getFloatFrequencyData
  # (https://developer.mozilla.org/en-US/docs/Web/API/AnalyserNode/getFloatFrequencyData).
  # It performs short-time Fourier transform (STFT) using a length-2048 Blackman
  # window. It opeartes on mono audio at the 44100-Hz sample rate.
  filepath = tf.keras.utils.get_file(
      filename,
      f'{TFJS_MODEL_ROOT}/speech-commands/conversion/{filename}',
      cache_subdir='model_maker',
      extract=True)
  model_path = _remove_suffix_if_possible(filepath, file_extension)
  return tf.keras.models.load_model(model_path)


def _load_tfjs_speech_command_model():
  """Download TFJS speech command model for fine-tune."""
  origin_root = f'{TFJS_MODEL_ROOT}/speech-commands/v0.3/browser_fft/18w'
  files_to_download = [
      'metadata.json', 'model.json', 'group1-shard1of2', 'group1-shard2of2'
  ]
  for filename in files_to_download:
    filepath = tf.keras.utils.get_file(
        filename,
        f'{origin_root}/{filename}',
        cache_subdir='model_maker/tfjs-sc-model')
  model_path = os.path.join(os.path.dirname(filepath), 'model.json')
  return model_util.load_tfjs_keras_model(model_path)


@mm_export('audio_classifier.BrowserFftSpec')
class BrowserFFTSpec(BaseSpec):
  """Model good at detecting speech commands, using Browser FFT spectrum."""

  EXPECTED_WAVEFORM_LENGTH = 44032

  # Information used to populate TFLite metadata.
  _MODEL_NAME = 'AudioClassifier'
  _MODEL_DESCRIPTION = ('Identify the most prominent type in the audio clip '
                        'from a known set of categories.')

  _MODEL_VERSION = 'v1'
  _MODEL_AUTHOR = 'TensorFlow Lite Model Maker'
  _MODEL_LICENSES = ('Apache License. Version 2.0 '
                     'http://www.apache.org/licenses/LICENSE-2.0.')

  _SAMPLE_RATE = 44100
  _CHANNELS = 1

  _INPUT_NAME = 'audio_clip'
  _INPUT_DESCRIPTION = 'Input audio clip to be classified.'

  _OUTPUT_NAME = 'probability'
  _OUTPUT_DESCRIPTION = 'Scores of the labels respectively.'

  def __init__(self, model_dir=None, strategy=None):
    """Initialize a new instance for BrowserFFT spec.

    Args:
      model_dir: The location to save the model checkpoint files.
      strategy: An instance of TF distribute strategy. If none, it will use the
        default strategy (either SingleDeviceStrategy or the current scoped
        strategy.
    """
    super(BrowserFFTSpec, self).__init__(model_dir, strategy)
    self._preprocess_model = _load_browser_fft_preprocess_model()
    self._tfjs_sc_model = _load_tfjs_speech_command_model()

  @property
  def target_sample_rate(self):
    return 44100

  @tf.function(input_signature=[
      tf.TensorSpec(shape=[None], dtype=tf.float32),
      tf.TensorSpec([], dtype=tf.int32)
  ])
  def _ensure_length(self, wav, unused_label):
    return len(wav) >= self.EXPECTED_WAVEFORM_LENGTH

  @tf.function(input_signature=[
      tf.TensorSpec(shape=[None], dtype=tf.float32),
      tf.TensorSpec([], dtype=tf.int32)
  ])
  def _split(self, wav, label):
    """Split the long audio samples into multiple trunks."""
    # wav shape: (audio_samples, )
    chunks = tf.math.floordiv(len(wav), self.EXPECTED_WAVEFORM_LENGTH)
    unused = tf.math.floormod(len(wav), self.EXPECTED_WAVEFORM_LENGTH)
    # Drop unused data
    wav = wav[:len(wav) - unused]
    # Split the audio sample into multiple chunks
    wav = tf.reshape(wav, (chunks, 1, self.EXPECTED_WAVEFORM_LENGTH))

    return wav, tf.repeat(tf.expand_dims(label, 0), len(wav))

  @tf.function(input_signature=[
      tf.TensorSpec(shape=[1, EXPECTED_WAVEFORM_LENGTH], dtype=tf.float32),
      tf.TensorSpec([], dtype=tf.int32)
  ])
  def _preprocess(self, x, label):
    """Preprocess the dataset to extract the spectrum."""
    # Add small Gaussian noise to the input x
    # to solve the potential "nan" problem of the preprocess_model.
    x = x + 1e-05 * tf.random.normal(x.shape)
    # x has shape (1, EXPECTED_WAVEFORM_LENGTH)
    spectrum = self._preprocess_model(x)
    # y has shape (1, embedding_len)
    spectrum = tf.squeeze(spectrum, axis=0)
    # y has shape (embedding_len,)
    return spectrum, label

  def preprocess_ds(self, ds, is_training=False, cache_fn=None):
    del is_training

    autotune = tf.data.AUTOTUNE
    ds = ds.filter(self._ensure_length)
    ds = ds.map(self._split, num_parallel_calls=autotune).unbatch()
    ds = ds.map(self._preprocess, num_parallel_calls=autotune)
    if cache_fn:
      ds = cache_fn(ds)
    return ds

  def create_model(self, num_classes, train_whole_model=False):
    if num_classes <= 1:
      raise ValueError(
          'AudioClassifier expects `num_classes` to be greater than 1')
    model = tf.keras.Sequential()
    for layer in self._tfjs_sc_model.layers[:-1]:
      model.add(layer)
    model.add(
        tf.keras.layers.Dense(
            name='classification_head', units=num_classes,
            activation='softmax'))
    if not train_whole_model:
      # Freeze all but the last layer of the model. The last layer will be
      # fine-tuned during transfer learning.
      for layer in model.layers[:-1]:
        layer.trainable = False
    return model

  def run_classifier(self, model, epochs, train_ds, validation_ds, **kwargs):
    model.compile(
        optimizer='adam', loss='categorical_crossentropy', metrics=['acc'])

    hist = model.fit(
        train_ds, validation_data=validation_ds, epochs=epochs, **kwargs)
    return hist

  def create_serving_model(self, training_model):
    """Create a model for serving."""
    combined = tf.keras.Sequential()
    combined.add(self._preprocess_model)
    combined.add(training_model)
    # Build the model.
    combined.build([None, self.EXPECTED_WAVEFORM_LENGTH])
    return combined

  def _export_metadata(self, tflite_filepath, index_to_label,
                       export_metadata_json_file):
    """Export TFLite metadata."""
    with MetadataWriter(
        tflite_filepath,
        name=self._MODEL_NAME,
        description=self._MODEL_DESCRIPTION,
        version=self._MODEL_VERSION,
        author=self._MODEL_AUTHOR,
        licenses=self._MODEL_LICENSES) as writer:
      writer.add_input(
          name=self._INPUT_NAME,
          description=self._INPUT_DESCRIPTION,
          sample_rate=self._SAMPLE_RATE,
          channels=self._CHANNELS)

      writer.add_output(
          labels=index_to_label,
          name=self._OUTPUT_NAME,
          description=self._OUTPUT_DESCRIPTION)

      json_filepath = (os.path.splitext(tflite_filepath)[0] +
                       '.json') if export_metadata_json_file else None
      writer.save(tflite_filepath, json_filepath)

  def export_tflite(self,
                    model,
                    tflite_filepath,
                    with_metadata=True,
                    export_metadata_json_file=True,
                    index_to_label=None,
                    quantization_config=None):
    """Converts the retrained model to tflite format and saves it.

    This method overrides the default `CustomModel._export_tflite` method, and
    include the pre-processing in the exported TFLite library since support
    library can't handle audio tasks yet.

    Args:
      model: An instance of the keras classification model to be exported.
      tflite_filepath: File path to save tflite model.
      with_metadata: Whether the output tflite model contains metadata.
      export_metadata_json_file: Whether to export metadata in json file. If
        True, export the metadata in the same directory as tflite model.Used
        only if `with_metadata` is True.
      index_to_label: A list that map from index to label class name.
      quantization_config: Configuration for post-training quantization.
    """
    combined = self.create_serving_model(model)

    # Sets batch size from None to 1 when converting to tflite.
    model_util.set_batch_size(model, batch_size=1)

    model_util.export_tflite(
        combined, tflite_filepath, quantization_config=quantization_config)

    # Sets batch size back to None to support retraining later.
    model_util.set_batch_size(model, batch_size=None)

    if with_metadata:
      if not ENABLE_METADATA:
        print('Writing Metadata is not support in the installed tflite-support '
              'version. Please use tflite-support >= 0.2.*')
      else:
        self._export_metadata(tflite_filepath, index_to_label,
                              export_metadata_json_file)


@mm_export('audio_classifier.YamNetSpec')
class YAMNetSpec(BaseSpec):
  """Model good at detecting environmental sounds, using YAMNet embedding."""

  EXPECTED_WAVEFORM_LENGTH = 15600  # effectively 0.975s
  EMBEDDING_SIZE = 1024

  # Information used to populate TFLite metadata.
  _MODEL_NAME = 'yamnet/classification'
  _MODEL_DESCRIPTION = 'Recognizes sound events'
  _MODEL_VERSION = 'v1'
  _MODEL_AUTHOR = 'TensorFlow Lite Model Maker'
  _MODEL_LICENSES = ('Apache License. Version 2.0 '
                     'http://www.apache.org/licenses/LICENSE-2.0.')

  _SAMPLE_RATE = 16000
  _CHANNELS = 1

  _INPUT_NAME = 'audio_clip'
  _INPUT_DESCRIPTION = 'Input audio clip to be classified.'

  _YAMNET_OUTPUT_NAME = 'yamnet'
  _YAMNET_OUTPUT_DESCRIPTION = ('Scores in range 0..1.0 for each of the 521 '
                                'output classes.')

  _CUSTOM_OUTPUT_NAME = 'custom'
  _CUSTOM_OUTPUT_DESCRIPTION = (
      'Scores in range 0..1.0 for each output classes.')

  def __init__(
      self,
      model_dir: None = None,
      strategy: None = None,
      yamnet_model_handle='https://tfhub.dev/google/yamnet/1',
      frame_length=EXPECTED_WAVEFORM_LENGTH,  # Window size 0.975 s
      frame_step=EXPECTED_WAVEFORM_LENGTH // 2,  # Hop of 0.975 /2 s
      keep_yamnet_and_custom_heads=True):
    """Initialize a new instance for YAMNet spec.

    Args:
      model_dir: The location to save the model checkpoint files.
      strategy: An instance of TF distribute strategy. If none, it will use the
        default strategy (either SingleDeviceStrategy or the current scoped
        strategy.
      yamnet_model_handle: Path of the TFHub model for retrining.
      frame_length: The number of samples in each audio frame. If the audio file
        is shorter than `frame_length`, then the audio file will be ignored.
      frame_step: The number of samples between two audio frames. This value
        should be smaller than `frame_length`, otherwise some samples will be
        ignored.
      keep_yamnet_and_custom_heads: Boolean, decides if the final TFLite model
        contains both YAMNet and custom trained classification heads. When set
        to False, only the trained custom head will be preserved.
    """
    super(YAMNetSpec, self).__init__(model_dir, strategy)
    self._yamnet_model_handle = yamnet_model_handle
    self._yamnet_model = hub.load(yamnet_model_handle)
    self._frame_length = frame_length
    self._frame_step = frame_step
    self._keep_yamnet_and_custom_heads = keep_yamnet_and_custom_heads

  @property
  def target_sample_rate(self):
    return self._SAMPLE_RATE

  def create_model(self, num_classes, train_whole_model=False):
    model = tf.keras.Sequential([
        tf.keras.layers.InputLayer(
            input_shape=(YAMNetSpec.EMBEDDING_SIZE),
            dtype=tf.float32,
            name='embedding'),
        tf.keras.layers.Dense(
            num_classes, name='classification_head', activation='softmax')
    ])
    return model

  def run_classifier(self, model, epochs, train_ds, validation_ds, **kwargs):
    model.compile(
        optimizer='adam', loss='categorical_crossentropy', metrics=['acc'])

    hist = model.fit(
        train_ds, validation_data=validation_ds, epochs=epochs, **kwargs)
    return hist

  # Annotate the TF function with input_signature to avoid re-tracing. Otherwise
  # the TF function gets retraced everytime the input shape is changed.
  # Check https://www.tensorflow.org/api_docs/python/tf/function#args_1 for more
  # information.
  @tf.function(input_signature=[
      tf.TensorSpec(shape=[None], dtype=tf.float32),
      tf.TensorSpec([], dtype=tf.int32)
  ])
  def _frame(self, wav, label):
    clips = tf.signal.frame(
        wav, frame_length=self._frame_length, frame_step=self._frame_step)
    batch_labels = tf.repeat(tf.expand_dims(label, 0), len(clips))

    return clips, batch_labels

  @tf.function(input_signature=[
      tf.TensorSpec(shape=[None], dtype=tf.float32),
      tf.TensorSpec([], dtype=tf.int32)
  ])
  def _extract_embedding(self, wav, label):
    _, embeddings, _ = self._yamnet_model(wav)  # (chunks, EMBEDDING_SIZE)
    embedding = tf.reduce_mean(embeddings, axis=0)
    return embedding, label

  @tf.function(input_signature=[
      tf.TensorSpec(shape=[EMBEDDING_SIZE], dtype=tf.float32),
      tf.TensorSpec([], dtype=tf.int32)
  ])
  def _add_noise(self, embedding, label):
    noise = tf.random.normal(
        embedding.shape, mean=0.0, stddev=.2, dtype=tf.dtypes.float32)
    return noise + embedding, label

  def preprocess_ds(self, ds, is_training=False, cache_fn=None):
    autotune = tf.data.AUTOTUNE
    ds = ds.map(self._frame, num_parallel_calls=autotune).unbatch()
    ds = ds.map(self._extract_embedding, num_parallel_calls=autotune)

    # Cache intermediate results right before data augmentation.
    if cache_fn:
      ds = cache_fn(ds)

    if is_training:
      ds = ds.map(self._add_noise, num_parallel_calls=autotune)
    return ds

  def _yamnet_labels(self):
    class_map_path = self._yamnet_model.class_map_path().numpy()
    class_map_csv_text = tf.io.read_file(class_map_path).numpy().decode('utf-8')
    class_map_csv = io.StringIO(class_map_csv_text)
    class_names = [
        display_name for (class_index, mid,
                          display_name) in csv.reader(class_map_csv)
    ]
    class_names = class_names[1:]  # Skip CSV header
    return class_names

  def _export_metadata(self, tflite_filepath, index_to_label,
                       export_metadata_json_file):
    """Export TFLite metadata."""
    with MetadataWriter(
        tflite_filepath,
        name=self._MODEL_NAME,
        description=self._MODEL_DESCRIPTION,
        version=self._MODEL_VERSION,
        author=self._MODEL_AUTHOR,
        licenses=self._MODEL_LICENSES) as writer:
      writer.add_input(
          name=self._INPUT_NAME,
          description=self._INPUT_DESCRIPTION,
          sample_rate=self._SAMPLE_RATE,
          channels=self._CHANNELS)

      if self._keep_yamnet_and_custom_heads:
        writer.add_output(
            labels=self._yamnet_labels(),
            name=self._YAMNET_OUTPUT_NAME,
            description=self._YAMNET_OUTPUT_DESCRIPTION)

      writer.add_output(
          labels=index_to_label,
          name=self._CUSTOM_OUTPUT_NAME,
          description=self._CUSTOM_OUTPUT_DESCRIPTION)

      json_filepath = (os.path.splitext(tflite_filepath)[0] +
                       '.json') if export_metadata_json_file else None
      writer.save(tflite_filepath, json_filepath)

  def create_serving_model(self, training_model):
    """Create a model for serving."""
    embedding_extraction_layer = hub.KerasLayer(
        self._yamnet_model_handle, trainable=False)
    keras_input = tf.keras.Input(
        shape=(YAMNetSpec.EXPECTED_WAVEFORM_LENGTH,),
        dtype=tf.float32,
        name='audio')  # (1, wav)
    reshaped_input = tf.reshape(keras_input,
                                (YAMNetSpec.EXPECTED_WAVEFORM_LENGTH,))  # (wav)

    scores, embeddings, _ = embedding_extraction_layer(reshaped_input)
    serving_outputs = training_model(embeddings)

    if self._keep_yamnet_and_custom_heads:
      serving_model = tf.keras.Model(keras_input, [scores, serving_outputs])
    else:
      serving_model = tf.keras.Model(keras_input, serving_outputs)

    return serving_model

  def export_tflite(self,
                    model,
                    tflite_filepath,
                    with_metadata=True,
                    export_metadata_json_file=True,
                    index_to_label=None,
                    quantization_config=None):
    """Converts the retrained model to tflite format and saves it.

    This method overrides the default `CustomModel._export_tflite` method, and
    include the spectrom extraction in the model.

    The exported model has input shape (1, number of wav samples)

    Args:
      model: An instance of the keras classification model to be exported.
      tflite_filepath: File path to save tflite model.
      with_metadata: Whether the output tflite model contains metadata.
      export_metadata_json_file: Whether to export metadata in json file. If
        True, export the metadata in the same directory as tflite model. Used
        only if `with_metadata` is True.
      index_to_label: A list that map from index to label class name.
      quantization_config: Configuration for post-training quantization.
    """
    serving_model = self.create_serving_model(model)

    # TODO(b/164229433): Remove SELECT_TF_OPS once changes in the bug are
    # released.
    model_util.export_tflite(
        serving_model, tflite_filepath, quantization_config=quantization_config)

    if with_metadata:
      if not ENABLE_METADATA:
        print('Writing Metadata is not support in the current tflite-support '
              'version. Please use tflite-support >= 0.2.*')
      else:
        self._export_metadata(tflite_filepath, index_to_label,
                              export_metadata_json_file)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from tensorflowjs import version


# File name for the indexing JSON file in an artifact directory.
ARTIFACT_MODEL_JSON_FILE_NAME = 'model.json'

# JSON string keys for fields of the indexing JSON.
ARTIFACT_MODEL_TOPOLOGY_KEY = 'modelTopology'
ARTIFACT_MODEL_INITIALIZER = 'modelInitializer'
ARTIFACT_WEIGHTS_MANIFEST_KEY = 'weightsManifest'

FORMAT_KEY = 'format'
TFJS_GRAPH_MODEL_FORMAT = 'graph-model'
TFJS_LAYERS_MODEL_FORMAT = 'layers-model'

GENERATED_BY_KEY = 'generatedBy'
CONVERTED_BY_KEY = 'convertedBy'

SIGNATURE_KEY = 'signature'
USER_DEFINED_METADATA_KEY = 'userDefinedMetadata'

# Model formats.
KERAS_SAVED_MODEL = 'keras_saved_model'
KERAS_MODEL = 'keras'
TF_SAVED_MODEL = 'tf_saved_model'
TF_HUB_MODEL = 'tf_hub'
TFJS_GRAPH_MODEL = 'tfjs_graph_model'
TFJS_LAYERS_MODEL = 'tfjs_layers_model'
TF_FROZEN_MODEL = 'tf_frozen_model'

# CLI argument strings.
INPUT_PATH = 'input_path'
OUTPUT_PATH = 'output_path'
INPUT_FORMAT = 'input_format'
OUTPUT_FORMAT = 'output_format'
OUTPUT_NODE = 'output_node_names'
SIGNATURE_NAME = 'signature_name'
SAVED_MODEL_TAGS = 'saved_model_tags'
QUANTIZATION_BYTES = 'quantization_bytes'
QUANTIZATION_TYPE_FLOAT16 = 'quantize_float16'
QUANTIZATION_TYPE_UINT8 = 'quantize_uint8'
QUANTIZATION_TYPE_UINT16 = 'quantize_uint16'
SPLIT_WEIGHTS_BY_LAYER = 'split_weights_by_layer'
VERSION = 'version'
SKIP_OP_CHECK = 'skip_op_check'
STRIP_DEBUG_OPS = 'strip_debug_ops'
WEIGHT_SHARD_SIZE_BYTES = 'weight_shard_size_bytes'
CONTROL_FLOW_V2 = 'control_flow_v2'
EXPERIMENTS = 'experiments'
METADATA = 'metadata'

def get_converted_by():
  """Get the convertedBy string for storage in model artifacts."""
  return 'TensorFlow.js Converter v%s' % version.version



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handles both V1 and V2 modules."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow_hub as hub


class HubKerasLayerV1V2(hub.KerasLayer):
  """Class to loads TF v1 and TF v1 hub modules that could be fine-tuned.

  Since TF v1 modules couldn't be retrained in hub.KerasLayer. This class
  provides a workaround for retraining the whole tf1 model in tf2. In
  particular, it extract self._func._self_unconditional_checkpoint_dependencies
  into trainable variable in tf1.

  Doesn't update moving-mean/moving-variance for BatchNormalization during
  fine-tuning.
  """

  def _setup_layer(self, trainable=False, **kwargs):
    if self._is_hub_module_v1:
      self._setup_layer_v1(trainable, **kwargs)
    else:
      # call _setup_layer from the base class for v2.
      super(HubKerasLayerV1V2, self)._setup_layer(trainable, **kwargs)

  def _check_trainability(self):
    if self._is_hub_module_v1:
      self._check_trainability_v1()
    else:
      # call _check_trainability from the base class for v2.
      super(HubKerasLayerV1V2, self)._check_trainability()

  def _setup_layer_v1(self, trainable=False, **kwargs):
    """Constructs keras layer with relevant weights and losses."""
    # Initialize an empty layer, then add_weight() etc. as needed.
    super(hub.KerasLayer, self).__init__(trainable=trainable, **kwargs)

    if not self._is_hub_module_v1:
      raise ValueError(
          'Only supports to set up v1 hub module in this function.')

    # v2 trainable_variable:
    if hasattr(self._func, 'trainable_variables'):
      for v in self._func.trainable_variables:
        self._add_existing_weight(v, trainable=True)
      trainable_variables = {id(v) for v in self._func.trainable_variables}
    else:
      trainable_variables = set()

    if not hasattr(self._func, '_self_unconditional_checkpoint_dependencies'):
      raise ValueError('_func doesn\'t contains attribute '
                       '_self_unconditional_checkpoint_dependencies.')
    dependencies = self._func._self_unconditional_checkpoint_dependencies  # pylint: disable=protected-access

    # Adds trainable variables.
    for dep in dependencies:
      if dep.name == 'variables':
        for v in dep.ref:
          if id(v) not in trainable_variables:
            self._add_existing_weight(v, trainable=True)
            trainable_variables.add(id(v))

    # Adds non-trainable variables.
    if hasattr(self._func, 'variables'):
      for v in self._func.variables:
        if id(v) not in trainable_variables:
          self._add_existing_weight(v, trainable=False)

    # Forward the callable's regularization losses (if any).
    if hasattr(self._func, 'regularization_losses'):
      for l in self._func.regularization_losses:
        if not callable(l):
          raise ValueError(
              'hub.KerasLayer(obj) expects obj.regularization_losses to be an '
              'iterable of callables, each returning a scalar loss term.')
        self.add_loss(self._call_loss_if_trainable(l))  # Supports callables.

  def _check_trainability_v1(self):
    """"Ignores trainability checks for V1."""
    if self._is_hub_module_v1:
      return  # Nothing to do.



# Lint as: python3
#   Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Dotproduct similarity layer."""
from typing import Optional

import tensorflow as tf


class DotProductSimilarity(tf.keras.layers.Layer):
  """Layer to comput dotproduct similarities for context/label embedding.

    The top_k is an integer to represent top_k ids to compute among label ids.
    if top_k is None, top_k computation will be ignored.
  """

  def call(self,
           context_embeddings: tf.Tensor,
           label_embeddings: tf.Tensor,
           top_k: Optional[int] = None):
    """Generate dotproduct similarity matrix and top values/indices.

    Args:
      context_embeddings: Context embeddings generated with input context
        sequence. The shape of tensor should be [batch_size, embedding_dim] for
        training mode, and [1, embedding_dim] for inference mode.
      label_embeddings: Label embeddings generated for candidate label items.
        The shape of tensor should be [num_items, embedding_dim].
      top_k: The number of top values to compute. If it's not set, top values
        and indices will not be computed.

    Returns:
      Dotproduct similarity matrix with shape [num_label_items, 1] and top
      values/indices if top_k is set.
    """
    if tf.keras.backend.ndim(label_embeddings) == 3:
      label_embeddings = tf.squeeze(label_embeddings, 1)
    dotproduct = tf.matmul(
        context_embeddings, label_embeddings, transpose_b=True)
    if top_k:
      top_values, top_indices = tf.math.top_k(
          tf.squeeze(dotproduct), top_k, sorted=True)
      top_ids = tf.identity(top_indices, name='top_prediction_ids')
      top_scores = tf.identity(
          tf.math.sigmoid(top_values), name='top_prediction_scores')
      return [dotproduct, top_ids, top_scores]
    else:
      return [dotproduct]



# Copyright 2020 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""A demo script to show to train a segmentation model."""
from absl import app
from absl import logging
import tensorflow as tf
import tensorflow_datasets as tfds

from tensorflow_examples.lite.model_maker.third_party.efficientdet import hparams_config
from tensorflow_examples.lite.model_maker.third_party.efficientdet.keras import efficientdet_keras


def create_mask(pred_mask):
  pred_mask = tf.argmax(pred_mask, axis=-1)
  pred_mask = pred_mask[..., tf.newaxis]
  return pred_mask[0]


def normalize(input_image, input_mask):
  input_image = tf.cast(input_image, tf.float32) / 255.0
  input_mask -= 1
  return input_image, input_mask


def load_image_train(datapoint):
  """Load images for training."""
  input_image = tf.image.resize(datapoint['image'], (512, 512))
  input_mask = tf.image.resize(datapoint['segmentation_mask'], (128, 128))

  if tf.random.uniform(()) > 0.5:
    input_image = tf.image.flip_left_right(input_image)
    input_mask = tf.image.flip_left_right(input_mask)

  input_image, input_mask = normalize(input_image, input_mask)

  return input_image, input_mask


def load_image_test(datapoint):
  input_image = tf.image.resize(datapoint['image'], (512, 512))
  input_mask = tf.image.resize(datapoint['segmentation_mask'], (128, 128))

  input_image, input_mask = normalize(input_image, input_mask)

  return input_image, input_mask


def main(_):
  dataset, info = tfds.load('oxford_iiit_pet:3.*.*', with_info=True)
  train_examples = info.splits['train'].num_examples
  batch_size = 8
  steps_per_epoch = train_examples // batch_size

  train = dataset['train'].map(
      load_image_train, num_parallel_calls=tf.data.experimental.AUTOTUNE)
  test = dataset['test'].map(load_image_test)

  train_dataset = train.cache().shuffle(1000).batch(batch_size).repeat()
  train_dataset = train_dataset.prefetch(
      buffer_size=tf.data.experimental.AUTOTUNE)
  test_dataset = test.batch(batch_size)
  config = hparams_config.get_efficientdet_config('efficientdet-d0')
  config.heads = ['segmentation']
  model = efficientdet_keras.EfficientDetNet(config=config)
  model.build((1, 512, 512, 3))
  model.compile(
      optimizer='adam',
      loss=tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True),
      metrics=['accuracy'])

  val_subsplits = 5
  val_steps = info.splits['test'].num_examples // batch_size // val_subsplits
  model.fit(
      train_dataset,
      epochs=20,
      steps_per_epoch=steps_per_epoch,
      validation_steps=val_steps,
      validation_data=test_dataset,
      callbacks=[])

  model.save_weights(
      './third_party/brain_automl/efficientdet/testdata/segmentation')

  print(create_mask(model(tf.ones((1, 512, 512, 3)), False)))


if __name__ == '__main__':
  logging.set_verbosity(logging.WARNING)
  app.run(main)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Resource management library."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os


def open_file(path):
  """Opens the file at given path, where path is relative to tensorflowjs/.

  Args:
    path: a string resource path relative to tensorflowjs/.

  Returns:
    An open file of that resource.

  Raises:
    IOError: If the path is not found, or the resource can't be opened.
  """
  path = os.path.join(os.path.dirname(os.path.abspath(__file__)), path)
  return open(path)


def list_dir(path):
  """List the files inside a dir where path is relative to tensorflowjs/.

  Args:
    path: a string path to a resource directory relative to tensorflowjs/.

  Returns:
    A list of files inside that directory.

  Raises:
    IOError: If the path is not found, or the resource can't be read.
  """
  path = os.path.join(os.path.dirname(os.path.abspath(__file__)), path)
  return os.listdir(path)



# Copyright 2020 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""COCO-style evaluation metrics.

Implements the interface of COCO API and metric_fn in tf.TPUEstimator.
COCO API: github.com/cocodataset/cocoapi/
"""
import json
import os
import sys
from absl import logging
import numpy as np
import tensorflow as tf

from tensorflow_examples.lite.model_maker.third_party.efficientdet.keras import label_util

try:
# pylint: disable=g-import-not-at-top
  from pycocotools.coco import COCO
  from pycocotools.cocoeval import COCOeval
# pylint: enable=g-import-not-at-top
except ImportError:
  COCO = None
  COCOeval = None


def block_print(log_level):
  """Disables print function when current logging level > log_level."""
  if tf.get_logger().getEffectiveLevel() > log_level:
    sys.stdout = open(os.devnull, 'w')


def enable_print(original_stdout):
  """Enables print function."""
  sys.stdout = original_stdout


class EvaluationMetric():
  """COCO evaluation metric class.

  This class cannot inherit from tf.keras.metrics.Metric due to numpy.
  """

  def __init__(self, filename=None, testdev_dir=None, label_map=None):
    """Constructs COCO evaluation class.

    The class provides the interface to metrics_fn in TPUEstimator. The
    _update_op() takes detections from each image and push them to
    self.detections. The _evaluate() loads a JSON file in COCO annotation format
    as the groundtruth and runs COCO evaluation.

    Args:
      filename: Ground truth JSON file name. If filename is None, use
        groundtruth data passed from the dataloader for evaluation. filename is
        ignored if testdev_dir is not None.
      testdev_dir: folder name for testdev data. If None, run eval without
        groundtruth, and filename will be ignored.
      label_map: a dict from id to class name. Used for per-class AP.
    """
    self.label_map = label_map
    self.filename = filename
    self.testdev_dir = testdev_dir
    self.metric_names = ['AP', 'AP50', 'AP75', 'APs', 'APm', 'APl', 'ARmax1',
                         'ARmax10', 'ARmax100', 'ARs', 'ARm', 'ARl']
    self.reset_states()

  def reset_states(self):
    """Reset COCO API object."""
    self.detections = []
    self.dataset = {
        'images': [],
        'annotations': [],
        'categories': []
    }
    self.image_id = 1
    self.annotation_id = 1
    self.category_ids = []
    self.metric_values = None

  def evaluate(self, log_level=tf.compat.v1.logging.INFO):
    """Evaluates with detections from all images with COCO API.

    Args:
      log_level: Logging lavel to print logs.
    Returns:
      coco_metric: float numpy array with shape [12] representing the
        coco-style evaluation metrics.
    Raises:
      ImportError: if the pip package `pycocotools` is not installed.
    """
    if COCO is None or COCOeval is None:
      message = ('You must install pycocotools (`pip install pycocotools`) '
                 '(see github repo at https://github.com/cocodataset/cocoapi) '
                 'for efficientdet/coco_metric to work.')
      raise ImportError(message)

    original_stdout = sys.stdout
    block_print(log_level)
    if self.filename:
      coco_gt = COCO(self.filename)
    else:
      coco_gt = COCO()
      coco_gt.dataset = self.dataset
      coco_gt.createIndex()
    enable_print(original_stdout)

    if self.testdev_dir:
      # Run on test-dev dataset.
      box_result_list = []
      for det in self.detections:
        box_result_list.append({
            'image_id': int(det[0]),
            'category_id': int(det[6]),
            'bbox': np.around(
                det[1:5].astype(np.float64), decimals=2).tolist(),
            'score': float(np.around(det[5], decimals=3)),
        })
      json.encoder.FLOAT_REPR = lambda o: format(o, '.3f')
      # Must be in the formst of 'detections_test-dev2017_xxx_results'.
      fname = 'detections_test-dev2017_test_results'
      output_path = os.path.join(self.testdev_dir, fname + '.json')
      logging.info('Writing output json file to: %s', output_path)
      with tf.io.gfile.GFile(output_path, 'w') as fid:
        json.dump(box_result_list, fid)
      return np.array([-1.], dtype=np.float32)
    else:
      # Run on validation dataset.
      block_print(log_level)
      detections = np.array(self.detections)
      image_ids = list(set(detections[:, 0]))
      coco_dt = coco_gt.loadRes(detections)
      coco_eval = COCOeval(coco_gt, coco_dt, iouType='bbox')
      coco_eval.params.imgIds = image_ids
      coco_eval.evaluate()
      coco_eval.accumulate()
      coco_eval.summarize()
      enable_print(original_stdout)
      coco_metrics = coco_eval.stats

      if self.label_map:
        # Get per_class AP, see pycocotools/cocoeval.py:334
        # TxRxKxAxM: iouThrs x recThrs x catIds x areaRng x maxDets
        # Use areaRng_id=0 ('all') and maxDets_id=-1 (200) in default
        precision = coco_eval.eval['precision'][:, :, :, 0, -1]
        # Ideally, label_map should match the eval set, but it is possible that
        # some classes has no data in the eval set.
        ap_perclass = [0] * max(precision.shape[-1], len(self.label_map))
        for c in range(precision.shape[-1]):  # iterate over all classes
          precision_c = precision[:, :, c]
          # Only consider values if > -1.
          precision_c = precision_c[precision_c > -1]
          ap_c = np.mean(precision_c) if precision_c.size else -1.
          ap_perclass[c] = ap_c
        coco_metrics = np.concatenate((coco_metrics, ap_perclass))

      # Return the concat normal and per-class AP.
      return np.array(coco_metrics, dtype=np.float32)

  def result(self, log_level=tf.compat.v1.logging.INFO):
    """Return the metric values (and compute it if needed)."""
    if self.metric_values is None:
      self.metric_values = self.evaluate(log_level)
    return self.metric_values

  def update_state(self, groundtruth_data, detections):
    """Update detection results and groundtruth data.

    Append detection results to self.detections to aggregate results from
    all validation set. The groundtruth_data is parsed and added into a
    dictionary with the same format as COCO dataset, which can be used for
    evaluation.

    Args:
      groundtruth_data: Groundtruth annotations in a tensor with each row
        representing [y1, x1, y2, x2, is_crowd, area, class].
      detections: Detection results in a tensor with each row representing
        [image_id, x, y, width, height, score, class].
    """
    for i, det in enumerate(detections):
      # Filter out detections with predicted class label = -1.
      indices = np.where(det[:, -1] > -1)[0]
      det = det[indices]
      if det.shape[0] == 0:
        continue
      # Append groundtruth annotations to create COCO dataset object.
      # Add images.
      image_id = det[0, 0]
      if image_id == -1:
        image_id = self.image_id
      det[:, 0] = image_id
      self.detections.extend(det)

      if not self.filename and not self.testdev_dir:
        # process groudtruth data only if filename is empty and no test_dev.
        self.dataset['images'].append({
            'id': int(image_id),
        })

        # Add annotations.
        indices = np.where(groundtruth_data[i, :, -1] > -1)[0]
        for data in groundtruth_data[i, indices]:
          box = data[0:4]
          is_crowd = data[4]
          area = (box[3] - box[1]) * (box[2] - box[0])
          category_id = data[6]
          if category_id < 0:
            break
          self.dataset['annotations'].append({
              'id': int(self.annotation_id),
              'image_id': int(image_id),
              'category_id': int(category_id),
              'bbox': [box[1], box[0], box[3] - box[1], box[2] - box[0]],
              'area': area,
              'iscrowd': int(is_crowd)
          })
          self.annotation_id += 1
          self.category_ids.append(category_id)

      self.image_id += 1

    if not self.filename:
      self.category_ids = list(set(self.category_ids))
      self.dataset['categories'] = [
          {'id': int(category_id)} for category_id in self.category_ids
      ]

  def estimator_metric_fn(self, detections, groundtruth_data):
    """Constructs the metric function for tf.TPUEstimator.

    For each metric, we return the evaluation op and an update op; the update op
    is shared across all metrics and simply appends the set of detections to the
    `self.detections` list. The metric op is invoked after all examples have
    been seen and computes the aggregate COCO metrics. Please find details API
    in: https://www.tensorflow.org/api_docs/python/tf/contrib/learn/MetricSpec

    Args:
      detections: Detection results in a tensor with each row representing
        [image_id, x, y, width, height, score, class]
      groundtruth_data: Groundtruth annotations in a tensor with each row
        representing [y1, x1, y2, x2, is_crowd, area, class].
    Returns:
      metrics_dict: A dictionary mapping from evaluation name to a tuple of
        operations (`metric_op`, `update_op`). `update_op` appends the
        detections for the metric to the `self.detections` list.
    """
    with tf.name_scope('coco_metric'):
      if self.testdev_dir:
        update_op = tf.numpy_function(self.update_state,
                                      [groundtruth_data, detections], [])
        metrics = tf.numpy_function(self.result, [], tf.float32)
        metrics_dict = {'AP': (metrics, update_op)}
        return metrics_dict
      else:
        update_op = tf.numpy_function(self.update_state,
                                      [groundtruth_data, detections], [])
        metrics = tf.numpy_function(self.result, [], tf.float32)
        metrics_dict = {}
        for i, name in enumerate(self.metric_names):
          metrics_dict[name] = (metrics[i], update_op)

        if self.label_map:
          # process per-class AP.
          label_map = label_util.get_label_map(self.label_map)
          for i, cid in enumerate(sorted(label_map.keys())):
            name = 'AP_/%s' % label_map[cid]
            metrics_dict[name] = (metrics[i + len(self.metric_names)],
                                  update_op)
        return metrics_dict



# Copyright 2020 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Training related libraries."""
import math
import os
import re
from absl import logging
import neural_structured_learning as nsl
import numpy as np

import tensorflow as tf
from tensorflow_addons.callbacks import AverageModelCheckpoint
import tensorflow_hub as hub

from tensorflow_examples.lite.model_maker.third_party.efficientdet import coco_metric
from tensorflow_examples.lite.model_maker.third_party.efficientdet import inference
from tensorflow_examples.lite.model_maker.third_party.efficientdet import iou_utils
from tensorflow_examples.lite.model_maker.third_party.efficientdet import utils
from tensorflow_examples.lite.model_maker.third_party.efficientdet.keras import anchors
from tensorflow_examples.lite.model_maker.third_party.efficientdet.keras import efficientdet_keras
from tensorflow_examples.lite.model_maker.third_party.efficientdet.keras import label_util
from tensorflow_examples.lite.model_maker.third_party.efficientdet.keras import postprocess
from tensorflow_examples.lite.model_maker.third_party.efficientdet.keras import util_keras
from tensorflow_model_optimization.python.core.sparsity.keras import pruning_wrapper


def _collect_prunable_layers(model):
  """Recursively collect the prunable layers in the model."""
  prunable_layers = []
  for layer in model._flatten_layers(recursive=False, include_self=False):  # pylint: disable=protected-access
    # A keras model may have other models as layers.
    if isinstance(layer, pruning_wrapper.PruneLowMagnitude):
      prunable_layers.append(layer)
    elif isinstance(layer, (tf.keras.Model, tf.keras.layers.Layer)):
      prunable_layers += _collect_prunable_layers(layer)

  return prunable_layers


class UpdatePruningStep(tf.keras.callbacks.Callback):
  """Keras callback which updates pruning wrappers with the optimizer step.

  This callback must be used when training a model which needs to be pruned. Not
  doing so will throw an error.
  Example:
  ```python
  model.fit(x, y,
      callbacks=[UpdatePruningStep()])
  ```
  """

  def __init__(self):
    super(UpdatePruningStep, self).__init__()
    self.prunable_layers = []

  def on_train_begin(self, logs=None):
    # Collect all the prunable layers in the model.
    self.prunable_layers = _collect_prunable_layers(self.model)
    self.step = tf.keras.backend.get_value(self.model.optimizer.iterations)

  def on_train_batch_begin(self, batch, logs=None):
    tuples = []

    for layer in self.prunable_layers:
      if layer.built:
        tuples.append((layer.pruning_step, self.step))

    tf.keras.backend.batch_set_value(tuples)
    self.step = self.step + 1

  def on_epoch_end(self, batch, logs=None):
    # At the end of every epoch, remask the weights. This ensures that when
    # the model is saved after completion, the weights represent mask*weights.
    weight_mask_ops = []

    for layer in self.prunable_layers:
      if layer.built and isinstance(layer, pruning_wrapper.PruneLowMagnitude):
        if tf.executing_eagerly():
          layer.pruning_obj.weight_mask_op()
        else:
          weight_mask_ops.append(layer.pruning_obj.weight_mask_op())

    tf.keras.backend.batch_get_value(weight_mask_ops)


class PruningSummaries(tf.keras.callbacks.TensorBoard):
  """A Keras callback for adding pruning summaries to tensorboard.

  Logs the sparsity(%) and threshold at a given iteration step.
  """

  def __init__(self, log_dir, update_freq='epoch', **kwargs):
    if not isinstance(log_dir, str) or not log_dir:
      raise ValueError(
          '`log_dir` must be a non-empty string. You passed `log_dir`='
          '{input}.'.format(input=log_dir))

    super().__init__(log_dir=log_dir, update_freq=update_freq, **kwargs)

    log_dir = self.log_dir + '/metrics'
    self._file_writer = tf.summary.create_file_writer(log_dir)

  def _log_pruning_metrics(self, logs, step):
    with self._file_writer.as_default():
      for name, value in logs.items():
        tf.summary.scalar(name, value, step=step)

      self._file_writer.flush()

  def on_epoch_begin(self, epoch, logs=None):
    if logs is not None:
      super().on_epoch_begin(epoch, logs)

    pruning_logs = {}
    params = []
    prunable_layers = _collect_prunable_layers(self.model)
    for layer in prunable_layers:
      for _, mask, threshold in layer.pruning_vars:
        params.append(mask)
        params.append(threshold)

    params.append(self.model.optimizer.iterations)

    values = tf.keras.backend.batch_get_value(params)
    iteration = values[-1]
    del values[-1]
    del params[-1]

    param_value_pairs = list(zip(params, values))

    for mask, mask_value in param_value_pairs[::2]:
      pruning_logs.update({mask.name + '/sparsity': 1 - np.mean(mask_value)})

    for threshold, threshold_value in param_value_pairs[1::2]:
      pruning_logs.update({threshold.name + '/threshold': threshold_value})

    self._log_pruning_metrics(pruning_logs, iteration)


def update_learning_rate_schedule_parameters(params):
  """Updates params that are related to the learning rate schedule."""
  batch_size = params['batch_size']
  # Learning rate is proportional to the batch size
  params['adjusted_learning_rate'] = (params['learning_rate'] * batch_size / 64)
  steps_per_epoch = params['steps_per_epoch']
  params['lr_warmup_step'] = int(params['lr_warmup_epoch'] * steps_per_epoch)
  params['first_lr_drop_step'] = int(params['first_lr_drop_epoch'] *
                                     steps_per_epoch)
  params['second_lr_drop_step'] = int(params['second_lr_drop_epoch'] *
                                      steps_per_epoch)
  params['total_steps'] = int(params['num_epochs'] * steps_per_epoch)


class StepwiseLrSchedule(tf.optimizers.schedules.LearningRateSchedule):
  """Stepwise learning rate schedule."""

  def __init__(self, adjusted_lr: float, lr_warmup_init: float,
               lr_warmup_step: int, first_lr_drop_step: int,
               second_lr_drop_step: int):
    """Build a StepwiseLrSchedule.

    Args:
      adjusted_lr: `float`, The initial learning rate.
      lr_warmup_init: `float`, The warm up learning rate.
      lr_warmup_step: `int`, The warm up step.
      first_lr_drop_step: `int`, First lr decay step.
      second_lr_drop_step: `int`, Second lr decay step.
    """
    super().__init__()
    logging.info('LR schedule method: stepwise')
    self.adjusted_lr = adjusted_lr
    self.lr_warmup_init = lr_warmup_init
    self.lr_warmup_step = lr_warmup_step
    self.first_lr_drop_step = first_lr_drop_step
    self.second_lr_drop_step = second_lr_drop_step

  def __call__(self, step):
    linear_warmup = (
        self.lr_warmup_init +
        (tf.cast(step, dtype=tf.float32) / self.lr_warmup_step *
         (self.adjusted_lr - self.lr_warmup_init)))
    learning_rate = tf.where(step < self.lr_warmup_step, linear_warmup,
                             self.adjusted_lr)
    lr_schedule = [[1.0, self.lr_warmup_step], [0.1, self.first_lr_drop_step],
                   [0.01, self.second_lr_drop_step]]
    for mult, start_global_step in lr_schedule:
      learning_rate = tf.where(step < start_global_step, learning_rate,
                               self.adjusted_lr * mult)
    return learning_rate


class CosineLrSchedule(tf.optimizers.schedules.LearningRateSchedule):
  """Cosine learning rate schedule."""

  def __init__(self, adjusted_lr: float, lr_warmup_init: float,
               lr_warmup_step: int, total_steps: int):
    """Build a CosineLrSchedule.

    Args:
      adjusted_lr: `float`, The initial learning rate.
      lr_warmup_init: `float`, The warm up learning rate.
      lr_warmup_step: `int`, The warm up step.
      total_steps: `int`, Total train steps.
    """
    super().__init__()
    logging.info('LR schedule method: cosine')
    self.adjusted_lr = adjusted_lr
    self.lr_warmup_init = lr_warmup_init
    self.lr_warmup_step = lr_warmup_step
    self.decay_steps = tf.cast(total_steps - lr_warmup_step, tf.float32)

  def __call__(self, step):
    linear_warmup = (
        self.lr_warmup_init +
        (tf.cast(step, dtype=tf.float32) / self.lr_warmup_step *
         (self.adjusted_lr - self.lr_warmup_init)))
    cosine_lr = 0.5 * self.adjusted_lr * (
        1 + tf.cos(math.pi * tf.cast(step, tf.float32) / self.decay_steps))
    return tf.where(step < self.lr_warmup_step, linear_warmup, cosine_lr)


class PolynomialLrSchedule(tf.optimizers.schedules.LearningRateSchedule):
  """Polynomial learning rate schedule."""

  def __init__(self, adjusted_lr: float, lr_warmup_init: float,
               lr_warmup_step: int, power: float, total_steps: int):
    """Build a PolynomialLrSchedule.

    Args:
      adjusted_lr: `float`, The initial learning rate.
      lr_warmup_init: `float`, The warm up learning rate.
      lr_warmup_step: `int`, The warm up step.
      power: `float`, power.
      total_steps: `int`, Total train steps.
    """
    super().__init__()
    logging.info('LR schedule method: polynomial')
    self.adjusted_lr = adjusted_lr
    self.lr_warmup_init = lr_warmup_init
    self.lr_warmup_step = lr_warmup_step
    self.power = power
    self.total_steps = total_steps

  def __call__(self, step):
    linear_warmup = (
        self.lr_warmup_init +
        (tf.cast(step, dtype=tf.float32) / self.lr_warmup_step *
         (self.adjusted_lr - self.lr_warmup_init)))
    polynomial_lr = self.adjusted_lr * tf.pow(
        1 - (tf.cast(step, dtype=tf.float32) / self.total_steps), self.power)
    return tf.where(step < self.lr_warmup_step, linear_warmup, polynomial_lr)


def learning_rate_schedule(params):
  """Learning rate schedule based on global step."""
  update_learning_rate_schedule_parameters(params)
  lr_decay_method = params['lr_decay_method']
  if lr_decay_method == 'stepwise':
    return StepwiseLrSchedule(params['adjusted_learning_rate'],
                              params['lr_warmup_init'],
                              params['lr_warmup_step'],
                              params['first_lr_drop_step'],
                              params['second_lr_drop_step'])

  if lr_decay_method == 'cosine':
    return CosineLrSchedule(params['adjusted_learning_rate'],
                            params['lr_warmup_init'], params['lr_warmup_step'],
                            params['total_steps'])

  if lr_decay_method == 'polynomial':
    return PolynomialLrSchedule(params['adjusted_learning_rate'],
                                params['lr_warmup_init'],
                                params['lr_warmup_step'],
                                params['poly_lr_power'], params['total_steps'])

  raise ValueError('unknown lr_decay_method: {}'.format(lr_decay_method))


def get_optimizer(params):
  """Get optimizer."""
  learning_rate = learning_rate_schedule(params)
  momentum = params['momentum']
  if params['optimizer'].lower() == 'sgd':
    logging.info('Use SGD optimizer')
    optimizer = tf.keras.optimizers.SGD(learning_rate, momentum=momentum)
  elif params['optimizer'].lower() == 'adam':
    logging.info('Use Adam optimizer')
    optimizer = tf.keras.optimizers.Adam(learning_rate, beta_1=momentum)
  else:
    raise ValueError('optimizers should be adam or sgd')

  moving_average_decay = params['moving_average_decay']
  if moving_average_decay:
    # TODO(tanmingxing): potentially add dynamic_decay for new tfa release.
    from tensorflow_addons import optimizers as tfa_optimizers  # pylint: disable=g-import-not-at-top
    optimizer = tfa_optimizers.MovingAverage(
        optimizer, average_decay=moving_average_decay, dynamic_decay=True)
  precision = utils.get_precision(params['strategy'], params['mixed_precision'])
  if precision == 'mixed_float16' and params['loss_scale']:
    optimizer = tf.keras.mixed_precision.LossScaleOptimizer(
        optimizer,
        initial_scale=params['loss_scale'])
  return optimizer


class COCOCallback(tf.keras.callbacks.Callback):
  """A utility for COCO eval callback."""

  def __init__(self, test_dataset, update_freq=None):
    super().__init__()
    self.test_dataset = test_dataset
    self.update_freq = update_freq

  def set_model(self, model: tf.keras.Model):
    self.model = model
    config = model.config
    self.config = config
    label_map = label_util.get_label_map(config.label_map)
    log_dir = os.path.join(config.model_dir, 'coco')
    self.file_writer = tf.summary.create_file_writer(log_dir)
    self.evaluator = coco_metric.EvaluationMetric(
        filename=config.val_json_file, label_map=label_map)

  @tf.function
  def _get_detections(self, images, labels):
    cls_outputs, box_outputs = self.model(images, training=False)
    detections = postprocess.generate_detections(self.config,
                                                 cls_outputs,
                                                 box_outputs,
                                                 labels['image_scales'],
                                                 labels['source_ids'])
    tf.numpy_function(self.evaluator.update_state,
                      [labels['groundtruth_data'],
                       postprocess.transform_detections(detections)], [])

  def on_epoch_end(self, epoch, logs=None):
    epoch += 1
    if self.update_freq and epoch % self.update_freq == 0:
      self.evaluator.reset_states()
      strategy = tf.distribute.get_strategy()
      count = self.config.eval_samples // self.config.batch_size
      dataset = self.test_dataset.take(count)
      dataset = strategy.experimental_distribute_dataset(dataset)
      for (images, labels) in dataset:
        strategy.run(self._get_detections, (images, labels))
      metrics = self.evaluator.result()
      eval_results = {}
      with self.file_writer.as_default(), tf.summary.record_if(True):
        for i, name in enumerate(self.evaluator.metric_names):
          tf.summary.scalar(name, metrics[i], step=epoch)
          eval_results[name] = metrics[i]
      return eval_results


class DisplayCallback(tf.keras.callbacks.Callback):
  """Display inference result callback."""

  def __init__(self, sample_image, output_dir, update_freq=None):
    super().__init__()
    image_file = tf.io.read_file(sample_image)
    self.sample_image = tf.expand_dims(
        tf.image.decode_jpeg(image_file, channels=3), axis=0)
    self.update_freq = update_freq
    self.output_dir = output_dir

  def set_model(self, model: tf.keras.Model):
    self.model = model
    config = model.config
    log_dir = os.path.join(config.model_dir, 'test_images')
    self.file_writer = tf.summary.create_file_writer(log_dir)
    self.min_score_thresh = config.nms_configs['score_thresh'] or 0.4
    self.max_boxes_to_draw = config.nms_configs['max_output_size'] or 100

  def on_train_batch_end(self, batch, logs=None):
    if self.update_freq and batch % self.update_freq == 0:
      self._draw_inference(batch)

  def _draw_inference(self, step):
    self.model.__class__ = efficientdet_keras.EfficientDetModel
    results = self.model(self.sample_image, training=False)
    boxes, scores, classes, valid_len = tf.nest.map_structure(np.array, results)
    length = valid_len[0]
    image = inference.visualize_image(
        self.sample_image[0],
        boxes[0][:length],
        classes[0].astype(np.int)[:length],
        scores[0][:length],
        label_map=self.model.config.label_map,
        min_score_thresh=self.min_score_thresh,
        max_boxes_to_draw=self.max_boxes_to_draw)

    with self.file_writer.as_default():
      tf.summary.image('Test image', tf.expand_dims(image, axis=0), step=step)
    self.model.__class__ = EfficientDetNetTrain


def get_callbacks(params, val_dataset=None):
  """Get callbacks for given params."""
  if params['moving_average_decay']:
    avg_callback = AverageModelCheckpoint(
        filepath=os.path.join(params['model_dir'], 'emackpt-{epoch:d}'),
        verbose=params['verbose'],
        save_freq=params['save_freq'],
        save_weights_only=True,
        update_weights=False)
    callbacks = [avg_callback]
  else:
    ckpt_callback = tf.keras.callbacks.ModelCheckpoint(
        os.path.join(params['model_dir'], 'ckpt-{epoch:d}'),
        verbose=params['verbose'],
        save_freq=params['save_freq'],
        save_weights_only=True)
    callbacks = [ckpt_callback]
  if params['model_optimizations'] and 'prune' in params['model_optimizations']:
    prune_callback = UpdatePruningStep()
    prune_summaries = PruningSummaries(
        log_dir=params['model_dir'],
        update_freq=params['steps_per_execution'],
        profile_batch=2 if params['profile'] else 0)
    callbacks += [prune_callback, prune_summaries]
  else:
    tb_callback = tf.keras.callbacks.TensorBoard(
        log_dir=params['model_dir'],
        update_freq=params['steps_per_execution'],
        profile_batch=2 if params['profile'] else 0)
    callbacks.append(tb_callback)
  if params.get('sample_image', None):
    display_callback = DisplayCallback(
        params.get('sample_image', None), params['model_dir'],
        params['img_summary_steps'])
    callbacks.append(display_callback)
  if (params.get('map_freq', None) and val_dataset and
      params['strategy'] != 'tpu'):
    coco_callback = COCOCallback(val_dataset, params['map_freq'])
    callbacks.append(coco_callback)
  return callbacks


class AdversarialLoss(tf.keras.losses.Loss):
  """Adversarial keras loss wrapper."""

  # TODO(fsx950223): WIP
  def __init__(self, adv_config, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.adv_config = adv_config
    self.model = None
    self.loss_fn = None
    self.tape = None
    self.built = False

  def build(self, model, loss_fn, tape):
    self.model = model
    self.loss_fn = loss_fn
    self.tape = tape
    self.built = True

  def call(self, features, y, y_pred, labeled_loss):
    return self.adv_config.multiplier * nsl.keras.adversarial_loss(
        features,
        y,
        self.model,
        self.loss_fn,
        predictions=y_pred,
        labeled_loss=self.labeled_loss,
        gradient_tape=self.tape)


class FocalLoss(tf.keras.losses.Loss):
  """Compute the focal loss between `logits` and the golden `target` values.

  Focal loss = -(1-pt)^gamma * log(pt)
  where pt is the probability of being classified to the true class.
  """

  def __init__(self, alpha, gamma, label_smoothing=0.0, **kwargs):
    """Initialize focal loss.

    Args:
      alpha: A float32 scalar multiplying alpha to the loss from positive
        examples and (1-alpha) to the loss from negative examples.
      gamma: A float32 scalar modulating loss from hard and easy examples.
      label_smoothing: Float in [0, 1]. If > `0` then smooth the labels.
      **kwargs: other params.
    """
    super().__init__(**kwargs)
    self.alpha = alpha
    self.gamma = gamma
    self.label_smoothing = label_smoothing

  @tf.autograph.experimental.do_not_convert
  def call(self, y, y_pred):
    """Compute focal loss for y and y_pred.

    Args:
      y: A tuple of (normalizer, y_true), where y_true is the target class.
      y_pred: A float32 tensor [batch, height_in, width_in, num_predictions].

    Returns:
      the focal loss.
    """
    normalizer, y_true = y
    alpha = tf.convert_to_tensor(self.alpha, dtype=y_pred.dtype)
    gamma = tf.convert_to_tensor(self.gamma, dtype=y_pred.dtype)

    # compute focal loss multipliers before label smoothing, such that it will
    # not blow up the loss.
    pred_prob = tf.sigmoid(y_pred)
    p_t = (y_true * pred_prob) + ((1 - y_true) * (1 - pred_prob))
    alpha_factor = y_true * alpha + (1 - y_true) * (1 - alpha)
    modulating_factor = (1.0 - p_t)**gamma

    # apply label smoothing for cross_entropy for each entry.
    y_true = y_true * (1.0 - self.label_smoothing) + 0.5 * self.label_smoothing
    ce = tf.nn.sigmoid_cross_entropy_with_logits(labels=y_true, logits=y_pred)

    # compute the final loss and return
    return alpha_factor * modulating_factor * ce / normalizer


class BoxLoss(tf.keras.losses.Loss):
  """L2 box regression loss."""

  def __init__(self, delta=0.1, **kwargs):
    """Initialize box loss.

    Args:
      delta: `float`, the point where the huber loss function changes from a
        quadratic to linear. It is typically around the mean value of regression
        target. For instances, the regression targets of 512x512 input with 6
        anchors on P3-P7 pyramid is about [0.1, 0.1, 0.2, 0.2].
      **kwargs: other params.
    """
    super().__init__(**kwargs)
    self.huber = tf.keras.losses.Huber(
        delta, reduction=tf.keras.losses.Reduction.NONE)

  @tf.autograph.experimental.do_not_convert
  def call(self, y_true, box_outputs):
    num_positives, box_targets = y_true
    normalizer = num_positives * 4.0
    mask = tf.cast(box_targets != 0.0, box_outputs.dtype)
    box_targets = tf.expand_dims(box_targets, axis=-1)
    box_outputs = tf.expand_dims(box_outputs, axis=-1)
    # TODO(fsx950223): remove cast when huber loss dtype is fixed.
    box_loss = tf.cast(self.huber(box_targets, box_outputs),
                       box_outputs.dtype) * mask
    box_loss = tf.reduce_sum(box_loss) / normalizer
    return box_loss


class BoxIouLoss(tf.keras.losses.Loss):
  """Box iou loss."""

  def __init__(self, iou_loss_type, min_level, max_level, num_scales,
               aspect_ratios, anchor_scale, image_size, **kwargs):
    super().__init__(**kwargs)
    self.iou_loss_type = iou_loss_type
    self.input_anchors = anchors.Anchors(min_level, max_level, num_scales,
                                         aspect_ratios, anchor_scale,
                                         image_size)

  @tf.autograph.experimental.do_not_convert
  def call(self, y_true, box_outputs):
    anchor_boxes = tf.tile(
        self.input_anchors.boxes,
        [box_outputs.shape[0] // self.input_anchors.boxes.shape[0], 1])
    num_positives, box_targets = y_true
    normalizer = num_positives * 4.0
    mask = tf.cast(box_targets != 0.0, box_outputs.dtype)
    box_outputs = anchors.decode_box_outputs(box_outputs, anchor_boxes) * mask
    box_targets = anchors.decode_box_outputs(box_targets, anchor_boxes) * mask
    box_iou_loss = iou_utils.iou_loss(box_outputs, box_targets,
                                      self.iou_loss_type)
    box_iou_loss = tf.reduce_sum(box_iou_loss) / normalizer
    return box_iou_loss


class EfficientDetNetTrain(efficientdet_keras.EfficientDetNet):
  """A customized trainer for EfficientDet.

  see https://www.tensorflow.org/guide/keras/customizing_what_happens_in_fit
  """

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    log_dir = os.path.join(self.config.model_dir, 'train_images')
    self.summary_writer = tf.summary.create_file_writer(log_dir)

  def _freeze_vars(self):
    if self.config.var_freeze_expr:
      return [
          v for v in self.trainable_variables
          if not re.match(self.config.var_freeze_expr, v.name)
      ]
    return self.trainable_variables

  def _reg_l2_loss(self, weight_decay, regex=r'.*(kernel|weight):0$'):
    """Return regularization l2 loss loss."""
    var_match = re.compile(regex)
    return weight_decay * tf.add_n([
        tf.nn.l2_loss(v) for v in self._freeze_vars() if var_match.match(v.name)
    ])

  def _detection_loss(self, cls_outputs, box_outputs, labels, loss_vals):
    """Computes total detection loss.

    Computes total detection loss including box and class loss from all levels.
    Args:
      cls_outputs: an OrderDict with keys representing levels and values
        representing logits in [batch_size, height, width, num_anchors].
      box_outputs: an OrderDict with keys representing levels and values
        representing box regression targets in [batch_size, height, width,
        num_anchors * 4].
      labels: the dictionary that returned from dataloader that includes
        groundtruth targets.
      loss_vals: A dict of loss values.

    Returns:
      total_loss: an integer tensor representing total loss reducing from
        class and box losses from all levels.
      cls_loss: an integer tensor representing total class loss.
      box_loss: an integer tensor representing total box regression loss.
      box_iou_loss: an integer tensor representing total box iou loss.
    """
    # Sum all positives in a batch for normalization and avoid zero
    # num_positives_sum, which would lead to inf loss during training
    dtype = cls_outputs[0].dtype
    num_positives_sum = tf.reduce_sum(labels['mean_num_positives']) + 1.0
    positives_momentum = self.config.positives_momentum or 0
    if positives_momentum > 0:
      # normalize the num_positive_examples for training stability.
      moving_normalizer_var = tf.Variable(
          0.0,
          name='moving_normalizer',
          dtype=dtype,
          synchronization=tf.VariableSynchronization.ON_READ,
          trainable=False,
          aggregation=tf.VariableAggregation.MEAN)
      num_positives_sum = tf.keras.backend.moving_average_update(
          moving_normalizer_var,
          num_positives_sum,
          momentum=self.config.positives_momentum)
    elif positives_momentum < 0:
      num_positives_sum = utils.cross_replica_mean(num_positives_sum)
    num_positives_sum = tf.cast(num_positives_sum, dtype)
    levels = range(len(cls_outputs))
    cls_losses = []
    box_losses = []
    for level in levels:
      # Onehot encoding for classification labels.
      cls_targets_at_level = tf.one_hot(
          labels['cls_targets_%d' % (level + self.config.min_level)],
          self.config.num_classes,
          dtype=dtype)

      if self.config.data_format == 'channels_first':
        bs, _, width, height, _ = cls_targets_at_level.get_shape().as_list()
        cls_targets_at_level = tf.reshape(cls_targets_at_level,
                                          [bs, -1, width, height])
      else:
        bs, width, height, _, _ = cls_targets_at_level.get_shape().as_list()
        cls_targets_at_level = tf.reshape(cls_targets_at_level,
                                          [bs, width, height, -1])

      class_loss_layer = self.loss.get(FocalLoss.__name__, None)
      if class_loss_layer:
        cls_loss = class_loss_layer([num_positives_sum, cls_targets_at_level],
                                    cls_outputs[level])
        if self.config.data_format == 'channels_first':
          cls_loss = tf.reshape(
              cls_loss, [bs, -1, width, height, self.config.num_classes])
        else:
          cls_loss = tf.reshape(
              cls_loss, [bs, width, height, -1, self.config.num_classes])
        cls_loss *= tf.cast(
            tf.expand_dims(
                tf.not_equal(
                    labels['cls_targets_%d' % (level + self.config.min_level)],
                    -2), -1), dtype)
        cls_loss_sum = tf.reduce_sum(cls_loss)
        cls_losses.append(tf.cast(cls_loss_sum, dtype))

      if self.config.box_loss_weight and self.loss.get(BoxLoss.__name__, None):
        box_targets_at_level = (
            labels['box_targets_%d' % (level + self.config.min_level)])
        box_loss_layer = self.loss[BoxLoss.__name__]
        box_losses.append(
            box_loss_layer([num_positives_sum, box_targets_at_level],
                           box_outputs[level]))

    if self.config.iou_loss_type:
      box_outputs = tf.concat([tf.reshape(v, [-1, 4]) for v in box_outputs],
                              axis=0)
      box_targets = tf.concat([
          tf.reshape(labels['box_targets_%d' %
                            (level + self.config.min_level)], [-1, 4])
          for level in levels
      ],
                              axis=0)
      box_iou_loss_layer = self.loss[BoxIouLoss.__name__]
      box_iou_loss = box_iou_loss_layer([num_positives_sum, box_targets],
                                        box_outputs)
      loss_vals['box_iou_loss'] = box_iou_loss
    else:
      box_iou_loss = 0

    cls_loss = tf.add_n(cls_losses) if cls_losses else 0
    box_loss = tf.add_n(box_losses) if box_losses else 0
    total_loss = (
        cls_loss + self.config.box_loss_weight * box_loss +
        self.config.iou_loss_weight * box_iou_loss)
    loss_vals['det_loss'] = total_loss
    loss_vals['cls_loss'] = cls_loss
    loss_vals['box_loss'] = box_loss
    return total_loss

  def train_step(self, data):
    """Train step.

    Args:
      data: Tuple of (images, labels). Image tensor with shape [batch_size,
        height, width, 3]. The height and width are fixed and equal.Input labels
        in a dictionary. The labels include class targets and box targets which
        are dense label maps. The labels are generated from get_input_fn
        function in data/dataloader.py.

    Returns:
      A dict record loss info.
    """
    images, labels = data
    if self.config.img_summary_steps:
      with self.summary_writer.as_default():
        tf.summary.image('input_image', images)
    with tf.GradientTape() as tape:
      if len(self.config.heads) == 2:
        cls_outputs, box_outputs, seg_outputs = util_keras.fp16_to_fp32_nested(
            self(images, training=True))
        loss_dtype = cls_outputs[0].dtype
      elif 'object_detection' in self.config.heads:
        cls_outputs, box_outputs = util_keras.fp16_to_fp32_nested(
            self(images, training=True))
        loss_dtype = cls_outputs[0].dtype
      elif 'segmentation' in self.config.heads:
        seg_outputs, = util_keras.fp16_to_fp32_nested(
            self(images, training=True))
        loss_dtype = seg_outputs.dtype
      else:
        raise ValueError('No valid head found: {}'.format(self.config.heads))
      labels = util_keras.fp16_to_fp32_nested(labels)

      total_loss = 0
      loss_vals = {}
      if 'object_detection' in self.config.heads:
        det_loss = self._detection_loss(cls_outputs, box_outputs, labels,
                                        loss_vals)
        total_loss += det_loss
      if 'segmentation' in self.config.heads:
        seg_loss_layer = (
            self.loss[tf.keras.losses.SparseCategoricalCrossentropy.__name__])
        seg_loss = seg_loss_layer(labels['image_masks'], seg_outputs)
        total_loss += seg_loss
        loss_vals['seg_loss'] = seg_loss

      reg_l2_loss = self._reg_l2_loss(self.config.weight_decay)
      loss_vals['reg_l2_loss'] = reg_l2_loss
      total_loss += tf.cast(reg_l2_loss, loss_dtype)
      if isinstance(self.optimizer,
                    tf.keras.mixed_precision.LossScaleOptimizer):
        scaled_loss = self.optimizer.get_scaled_loss(total_loss)
        optimizer = self.optimizer.inner_optimizer
      else:
        scaled_loss = total_loss
        optimizer = self.optimizer
    loss_vals['loss'] = total_loss
    loss_vals['learning_rate'] = optimizer.learning_rate(optimizer.iterations)
    trainable_vars = self._freeze_vars()
    scaled_gradients = tape.gradient(scaled_loss, trainable_vars)
    if isinstance(self.optimizer,
                  tf.keras.mixed_precision.LossScaleOptimizer):
      gradients = self.optimizer.get_unscaled_gradients(scaled_gradients)
    else:
      gradients = scaled_gradients
    if self.config.clip_gradients_norm > 0:
      clip_norm = abs(self.config.clip_gradients_norm)
      gradients = [
          tf.clip_by_norm(g, clip_norm) if g is not None else None
          for g in gradients
      ]
      gradients, _ = tf.clip_by_global_norm(gradients, clip_norm)
      loss_vals['gradient_norm'] = tf.linalg.global_norm(gradients)
    self.optimizer.apply_gradients(zip(gradients, trainable_vars))
    return loss_vals

  def test_step(self, data):
    """Test step.

    Args:
      data: Tuple of (images, labels). Image tensor with shape [batch_size,
        height, width, 3]. The height and width are fixed and equal.Input labels
        in a dictionary. The labels include class targets and box targets which
        are dense label maps. The labels are generated from get_input_fn
        function in data/dataloader.py.

    Returns:
      A dict record loss info.
    """
    images, labels = data
    if len(self.config.heads) == 2:
      cls_outputs, box_outputs, seg_outputs = util_keras.fp16_to_fp32_nested(
          self(images, training=False))
      loss_dtype = cls_outputs[0].dtype
    elif 'object_detection' in self.config.heads:
      cls_outputs, box_outputs = util_keras.fp16_to_fp32_nested(
          self(images, training=False))
      loss_dtype = cls_outputs[0].dtype
    elif 'segmentation' in self.config.heads:
      seg_outputs, = util_keras.fp16_to_fp32_nested(
          self(images, training=False))
      loss_dtype = seg_outputs.dtype
    else:
      raise ValueError('No valid head found: {}'.format(self.config.heads))

    labels = util_keras.fp16_to_fp32_nested(labels)

    total_loss = 0
    loss_vals = {}
    if 'object_detection' in self.config.heads:
      det_loss = self._detection_loss(cls_outputs, box_outputs, labels,
                                      loss_vals)
      total_loss += det_loss
    if 'segmentation' in self.config.heads:
      seg_loss_layer = (
          self.loss[tf.keras.losses.SparseCategoricalCrossentropy.__name__])
      seg_loss = seg_loss_layer(labels['image_masks'], seg_outputs)
      total_loss += seg_loss
      loss_vals['seg_loss'] = seg_loss
    reg_l2_loss = self._reg_l2_loss(self.config.weight_decay)
    loss_vals['reg_l2_loss'] = reg_l2_loss
    loss_vals['loss'] = total_loss + tf.cast(reg_l2_loss, loss_dtype)
    return loss_vals


class EfficientDetNetTrainHub(EfficientDetNetTrain):
  """EfficientDetNetTrain for Hub module."""

  def __init__(self, config, hub_module_url, name=''):
    super(efficientdet_keras.EfficientDetNet, self).__init__(name=name)
    self.config = config
    self.hub_module_url = hub_module_url
    self.base_model = hub.KerasLayer(hub_module_url, trainable=True)

    # class/box output prediction network.
    num_anchors = len(config.aspect_ratios) * config.num_scales

    conv2d_layer = efficientdet_keras.ClassNet.conv2d_layer(
        config.separable_conv, config.data_format)
    self.classes = efficientdet_keras.ClassNet.classes_layer(
        conv2d_layer,
        config.num_classes,
        num_anchors,
        name='class_net/class-predict')

    self.boxes = efficientdet_keras.BoxNet.boxes_layer(
        config.separable_conv,
        num_anchors,
        config.data_format,
        name='box_net/box-predict')

    log_dir = os.path.join(self.config.model_dir, 'train_images')
    self.summary_writer = tf.summary.create_file_writer(log_dir)

  def call(self, inputs, training):
    cls_outputs, box_outputs = self.base_model(inputs, training=training)
    for i in range(self.config.max_level - self.config.min_level + 1):
      cls_outputs[i] = self.classes(cls_outputs[i])
      box_outputs[i] = self.boxes(box_outputs[i])
    return (cls_outputs, box_outputs)



# Copyright 2020 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Mosaic augmentation."""
import functools
from typing import Tuple
from absl import logging

import tensorflow as tf


class Mosaic:
  """Mosaic Augmentation class.

  1. Mosaic sub images will not be preserving aspect ratio of original images.
  2. Tested on static graphs and eager  execution.
  3. This Implementation of mosaic augmentation is tested in tf2.x.
  """

  def __init__(self,
               out_size: Tuple(int, int)=(680, 680),
               n_images: int = 4,
               _minimum_mosaic_image_dim: int = 25):
    """__init__.

    Args:
      out_size: output mosaic image size.
      n_images: number images to make mosaic
      _minimum_mosaic_image_dim: minimum percentage of out_size dimension
        should the mosaic be. i.e if out_size is (680,680) and
        _minimum_mosaic_image_dim is 25 , minimum mosaic sub images
        dimension will be 25 % of 680.
    """
    # TODO(someone) #MED #use n_images to build mosaic.
    self._n_images = n_images
    self._out_size = out_size
    self._minimum_mosaic_image_dim = _minimum_mosaic_image_dim
    assert (_minimum_mosaic_image_dim >
            0), "Minimum Mosaic image dimension should be above 0"

  @property
  def n_images(self) -> int:
    return self._n_images

  @property
  def out_size(self) -> int:
    return self._out_size

  def _mosaic_divide_points(self) -> Tuple(int, int):
    """Returns a  tuple of x and y which corresponds to mosaic divide points."""
    x_point = tf.random.uniform(
        shape=[1],
        minval=tf.cast(
            self.out_size[0] * (self._minimum_mosaic_image_dim / 100),
            tf.int32),
        maxval=tf.cast(
            self.out_size[0] * ((100 - self._minimum_mosaic_image_dim) / 100),
            tf.int32,
        ),
        dtype=tf.int32,
    )
    y_point = tf.random.uniform(
        shape=[1],
        minval=tf.cast(
            self.out_size[1] * (self._minimum_mosaic_image_dim / 100),
            tf.int32),
        maxval=tf.cast(
            self.out_size[1] * ((100 - self._minimum_mosaic_image_dim) / 100),
            tf.int32,
        ),
        dtype=tf.int32,
    )
    return x_point, y_point

  @staticmethod
  def _scale_box(box, image, mosaic_image):
    """scale boxes with mosaic sub image.

    Args:
      box: mosaic image box.
      image: original image.
      mosaic_image: mosaic sub image.

    Returns:
      Scaled bounding boxes.
    """
    return [
        box[0] * tf.shape(mosaic_image)[1] / tf.shape(image)[1],
        box[1] * tf.shape(mosaic_image)[0] / tf.shape(image)[0],
        box[2] * tf.shape(mosaic_image)[1] / tf.shape(image)[1],
        box[-1] * tf.shape(mosaic_image)[0] / tf.shape(image)[0],
    ]

  def _scale_images(self, images, mosaic_divide_points):
    """Scale Sub Images.

    Args:
      images: original single images to make mosaic.
      mosaic_divide_points: Points to build mosaic around on given output.

    Returns:
      A tuple of scaled Mosaic sub images.
    """
    x, y = mosaic_divide_points[0][0], mosaic_divide_points[1][0]
    mosaic_image_topleft = tf.image.resize(images[0], (x, y))
    mosaic_image_topright = tf.image.resize(images[1],
                                            (self.out_size[0] - x, y))
    mosaic_image_bottomleft = tf.image.resize(images[2],
                                              (x, self.out_size[1] - y))
    mosaic_image_bottomright = tf.image.resize(
        images[3], (self.out_size[0] - x, self.out_size[1] - y))
    return (
        mosaic_image_topleft,
        mosaic_image_topright,
        mosaic_image_bottomleft,
        mosaic_image_bottomright,
    )

  @tf.function
  def _mosaic(self, images, boxes, mosaic_divide_points):
    """Builds mosaic of provided images.

    Args:
      images: original single images to make mosaic.
      boxes: corresponding bounding boxes to images.
      mosaic_divide_points: Points to build mosaic around on given output size.

    Returns:
      A tuple of mosaic Image, Mosaic Boxes merged.
    """
    (
        mosaic_image_topleft,
        mosaic_image_topright,
        mosaic_image_bottomleft,
        mosaic_image_bottomright,
    ) = self._scale_images(images, mosaic_divide_points)

    #####################################################
    # Scale Boxes for TOP LEFT image.
    # Note: Below function is complex because of TF item assignment restriction.
    # Map_fn is replace with vectorized_map below for optimization purpose.
    mosaic_box_topleft = tf.transpose(
        tf.vectorized_map(
            functools.partial(
                self._scale_box,
                image=images[0],
                mosaic_image=mosaic_image_topleft),
            boxes[0],
        ))

    # Scale and Pad Boxes for TOP RIGHT image.

    mosaic_box_topright = tf.vectorized_map(
        functools.partial(
            self._scale_box,
            image=images[1],
            mosaic_image=mosaic_image_topright),
        boxes[1],
    )
    num_boxes = boxes[1].shape[0]
    idx_tp = tf.constant([[1], [3]])
    update_tp = [
        [tf.shape(mosaic_image_topleft)[0]] * num_boxes,
        [tf.shape(mosaic_image_topleft)[0]] * num_boxes,
    ]
    mosaic_box_topright = tf.transpose(
        tf.tensor_scatter_nd_add(mosaic_box_topright, idx_tp, update_tp))

    # Scale and Pad Boxes for BOTTOM LEFT image.

    mosaic_box_bottomleft = tf.vectorized_map(
        functools.partial(
            self._scale_box,
            image=images[2],
            mosaic_image=mosaic_image_bottomleft),
        boxes[2],
    )

    num_boxes = boxes[2].shape[0]
    idx_bl = tf.constant([[0], [2]])
    update_bl = [
        [tf.shape(mosaic_image_topleft)[1]] * num_boxes,
        [tf.shape(mosaic_image_topleft)[1]] * num_boxes,
    ]
    mosaic_box_bottomleft = tf.transpose(
        tf.tensor_scatter_nd_add(mosaic_box_bottomleft, idx_bl, update_bl))

    # Scale and Pad Boxes for BOTTOM RIGHT image.
    mosaic_box_bottomright = tf.vectorized_map(
        functools.partial(
            self._scale_box,
            image=images[3],
            mosaic_image=mosaic_image_bottomright),
        boxes[3],
    )

    num_boxes = boxes[3].shape[0]
    idx_br = tf.constant([[0], [2], [1], [3]])
    update_br = [
        [tf.shape(mosaic_image_topright)[1]] * num_boxes,
        [tf.shape(mosaic_image_topright)[1]] * num_boxes,
        [tf.shape(mosaic_image_bottomleft)[0]] * num_boxes,
        [tf.shape(mosaic_image_bottomleft)[0]] * num_boxes,
    ]
    mosaic_box_bottomright = tf.transpose(
        tf.tensor_scatter_nd_add(mosaic_box_bottomright, idx_br, update_br))

    # Gather mosaic_sub_images and boxes.
    mosaic_images = [
        mosaic_image_topleft,
        mosaic_image_topright,
        mosaic_image_bottomleft,
        mosaic_image_bottomright,
    ]
    mosaic_boxes = [
        mosaic_box_topleft,
        mosaic_box_topright,
        mosaic_box_bottomleft,
        mosaic_box_bottomright,
    ]

    return mosaic_images, mosaic_boxes

  def __call__(self, images, boxes):
    """Builds mosaic with given images, boxes."""
    if images.shape[0] != 4:
      err_msg = "Currently Exact 4 Images are supported by Mosaic Aug."
      logging.error(err_msg)
      raise Exception(err_msg)

    x, y = self._mosaic_divide_points()
    mosaic_sub_images, mosaic_boxes = self._mosaic(
        images, boxes, mosaic_divide_points=(x, y))

    upper_stack = tf.concat([mosaic_sub_images[0], mosaic_sub_images[1]],
                            axis=0)
    lower_stack = tf.concat([mosaic_sub_images[2], mosaic_sub_images[3]],
                            axis=0)
    mosaic_image = tf.concat([upper_stack, lower_stack], axis=1)
    return mosaic_image, mosaic_boxes
